package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.domain.Review;
import at.ac.uibk.gitsearch.domain.ReviewRating;
import at.ac.uibk.gitsearch.domain.User;
import at.ac.uibk.gitsearch.repository.jpa.ReviewRatingRepository;
import at.ac.uibk.gitsearch.repository.jpa.ReviewRepository;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider;
import at.ac.uibk.gitsearch.service.dto.ReviewDTO;
import at.ac.uibk.gitsearch.service.dto.ReviewRequest;
import at.ac.uibk.gitsearch.service.dto.StatisticsDTO;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.core.userdetails.UsernameNotFoundException;
import org.springframework.stereotype.Service;

@Service
public class ReviewService {

    private final Logger log = LoggerFactory.getLogger(ReviewService.class);

    @Autowired
    private UserService userService;

    @Autowired
    private ReviewRepository reviewRepository;

    @Autowired
    private ReviewRatingRepository reviewRatingRepository;

    @Autowired
    private StatisticsService statisticsService;

    @Autowired
    protected TokenProvider tokenProvider;

    @Autowired
    private MailService mailService;

    public Review createReview(ReviewRequest reviewRequest) throws UsernameNotFoundException {
        Review review = new Review();
        review.setResource(reviewRequest.getResource());
        review.setLink(reviewRequest.getLink());
        review.setResourceID(reviewRequest.getResourceID());
        User loggedInUser = userService.getUserWithAuthorities().orElse(null);
        if (loggedInUser == null) {
            throw new UsernameNotFoundException("User not found");
        }
        review.setRequestedBy(loggedInUser.getId());
        Set<ReviewRating> reviewRatings = new HashSet<>();

        reviewRequest
            .getUsers()
            .stream()
            .map(email -> userService.getUserByEmail(email).get())
            .forEach(user -> {
                ReviewRating reviewRating = new ReviewRating();
                reviewRating.setReview(review);
                reviewRating.setUser(user);
                reviewRating.setComment("");
                reviewRating.setStatus("Review In Progress");
                reviewRatings.add(reviewRating);
                review.setComments(reviewRatings);
                reviewRepository.save(review);
                reviewRatingRepository.save(reviewRating);
                mailService.sendEmail(
                    user.getEmail(),
                    "Review Updated",
                    "You have been selected to be a reviewer for " +
                    review.getResource() +
                    ". Please go to the sharing platform and review the resource.",
                    false,
                    false
                );
            });

        log.debug("REST request to create a review");
        return review;
    }

    public void updateStatus(Long id, String status) {
        log.debug("REST request to update a review status");
        Review review = reviewRepository.findById(id).get();
        review
            .getComments()
            .stream()
            .forEach(reviewRating -> {
                reviewRating.setStatus(status);
                reviewRatingRepository.save(reviewRating);
            });
    }

    public void updateReview(ReviewDTO review) {
        log.debug("REST request to update a review");
        Review reviewEntity = reviewRepository.findById(review.getId()).get();
        reviewEntity
            .getComments()
            .stream()
            .filter(reviewRating -> !review.getUsers().contains(reviewRating.getUser().getEmail()))
            .forEach(reviewRating -> reviewRatingRepository.delete(reviewRating));
        reviewEntity.setResource(review.getResource());
        Set<ReviewRating> reviewRatings = new HashSet<>();
        review
            .getUsers()
            .stream()
            .filter(email -> reviewEntity.getComments().stream().noneMatch(reviewRating -> reviewRating.getUser().getEmail().equals(email)))
            .forEach(email -> {
                ReviewRating reviewRating = new ReviewRating();
                reviewRating.setReview(reviewEntity);
                reviewRating.setUser(userService.getUserByEmail(email).get());
                reviewRating.setComment("");
                reviewRating.setStatus("Review In Progress");
                reviewRatings.add(reviewRating);
                reviewRatingRepository.save(reviewRating);
            });

        reviewEntity.setComments(reviewRatings);
        reviewRepository.save(reviewEntity);

        notifyReviewers(
            review.getId(),
            "You have been selected to be a reviewer for " +
            review.getResource() +
            ". Please go to the sharing platform and review the resource."
        );
    }

    public List<ReviewDTO> getAllReviews() {
        log.debug("REST request to get all reviews");
        List<Review> reviews = reviewRepository.findAll();
        return reviews
            .stream()
            .map(review -> {
                ReviewDTO temp = reviewToDTO(review);
                temp.setLink(review.getLink());
                return temp;
            })
            .collect(Collectors.toList());
    }

    public List<ReviewDTO> getAllReviewsWithStatistics() {
        log.debug("REST request to get all reviews with statistics");
        List<Review> reviews = reviewRepository.findAll();
        return reviews
            .stream()
            .map(review -> {
                ReviewDTO temp = reviewToDTO(review);
                temp.setLink(review.getLink());
                StatisticsDTO stats = statisticsService.findOneByExerciseID(review.getResourceID()).orElse(null);
                if (stats == null) {
                    temp.setBadgeRewarded(false);
                } else {
                    temp.setBadgeRewarded(stats.getBadgeRewarded());
                }
                return temp;
            })
            .collect(Collectors.toList());
    }

    public List<ReviewDTO> getReviewsByUser() {
        log.debug("REST request to get all reviews by user");
        User user = userService.getUserWithAuthorities().orElse(null);
        if (user == null) {
            return new ArrayList<>();
        }
        List<ReviewRating> reviewRating = reviewRatingRepository.findAllByUserId(user.getId());

        List<ReviewDTO> reviewRatingDTOs = reviewRating
            .stream()
            .map(reviewRating1 ->
                new ReviewDTO(
                    reviewRating1.getReview().getId(),
                    reviewRating1.getReview().getResource(),
                    reviewRating1.getReview().getLink(),
                    Arrays.asList(user.getEmail()),
                    Arrays.asList(reviewRating1.getStatus()),
                    Arrays.asList(reviewRating1.getComment()),
                    "ANNONYMIZED"
                )
            )
            .collect(Collectors.toList());

        List<Review> reviews = reviewRepository.findAllByRequestedBy(user.getId());
        List<ReviewDTO> reviewDTOs = reviews
            .stream()
            .map(review -> {
                ReviewDTO temp = reviewToDTO(review);
                temp.setLink(review.getLink());
                return temp;
            })
            .collect(Collectors.toList());
        reviewDTOs
            .stream()
            .forEach(review -> {
                review.setRequestedBy(user.getEmail());
                review.setUsers(review.getUsers().stream().map(x -> "ANNONYMIZED").collect(Collectors.toList()));
            });
        reviewDTOs.addAll(reviewRatingDTOs);
        return reviewDTOs;
    }

    public void updateReviewRatingByUser(ReviewDTO reviewDTO) throws UsernameNotFoundException, IllegalAccessException {
        log.debug("REST request to update a review by user");
        User user = userService.getUserWithAuthorities().orElse(null);
        if (user == null) {
            throw new UsernameNotFoundException("User not found, unable to update ReviewRating");
        }
        Review review = reviewRepository.findById(reviewDTO.getId()).get();
        if (review.getRequestedBy().equals(user.getId()) && !"DELETION REQUESTED".equals(reviewDTO.getStatus().get(0))) {
            throw new IllegalAccessException("User is not allowed to update ReviewRating");
        }
        if (review.getRequestedBy().equals(user.getId()) && "DELETION REQUESTED".equals(reviewDTO.getStatus().get(0))) {
            notifyReviewers(review.getId(), "The review for " + review.getResource() + " has been deleted by the user.");
            deleteReview(review.getId());
            return;
        }

        ReviewRating reviewRating = review
            .getComments()
            .stream()
            .filter(reviewRating1 -> reviewRating1.getUser().getEmail().equals(user.getEmail()))
            .findFirst()
            .get();
        reviewRating.setComment(reviewDTO.getComments().get(0));
        reviewRating.setStatus(reviewDTO.getStatus().get(0));
        reviewRatingRepository.save(reviewRating);
        mailService.sendEmail(
            userService.getUserById(review.getRequestedBy()).get().getEmail(),
            "New feedback for resource",
            "A reviewer has just updated you review request for resource: " + review.getResource(),
            false,
            false
        );
    }

    public void deleteReview(Long id) {
        log.debug("REST request to delete review: {}", id);
        Review review = reviewRepository.findById(id).get();
        StatisticsDTO statistics = statisticsService.findOneByExerciseID(review.getResourceID()).orElse(null);
        if (statistics != null) {
            statistics.setBadgeRewarded(false);
            statisticsService.save(statistics);
        }
        notifyReviewers(review.getId(), "The review for " + review.getResource() + " has been deleted.");
        reviewRatingRepository.deleteByReviewId(id);
        reviewRepository.deleteById(id);
    }

    public ReviewDTO reviewToDTO(Review review) {
        ReviewDTO reviewDTO = new ReviewDTO();
        reviewDTO.setResource(review.getResource());
        reviewDTO.setUsers(
            review.getComments().stream().map(reviewRating -> reviewRating.getUser().getEmail()).collect(Collectors.toList())
        );
        reviewDTO.setId(review.getId());
        reviewDTO.setStatus(review.getComments().stream().map(reviewRating -> reviewRating.getStatus()).collect(Collectors.toList()));
        reviewDTO.setComments(review.getComments().stream().map(reviewRating -> reviewRating.getComment()).collect(Collectors.toList()));
        return reviewDTO;
    }

    public ReviewDTO getReview(Long id) {
        Review review = reviewRepository.findById(id).get();
        return reviewToDTO(review);
    }

    public void rewardBadge(Long id) {
        Review review = reviewRepository.findById(id).get();
        updateStatus(review.getId(), "Review completed");
        statisticsService.rewardBadgeForStatisticsByExerciseID(review.getResourceID());
        notifyReviewers(
            review.getId(),
            "The review for " + review.getResource() + " has been completed and a badge is now being displayed when people search for it."
        );
    }

    public void requestReviewForExercise(ReviewRequest request) {
        String id = request.getResource();
        if (reviewRepository.findOneByResource(id).isPresent()) {
            log.warn("Review already exists for exercise: {}", id);
            return;
        }

        Review review = new Review();
        review.setResource(id);
        review.setLink(request.getLink());
        review.setResourceID(request.getResourceID());

        Optional<User> user = userService.getUserWithAuthorities();
        if (user.isPresent()) {
            review.setRequestedBy(user.get().getId());
            // reviewRating.setUser(user.get());
        } else {
            log.error("User not found");
            return;
        }
        Set<ReviewRating> reviewRatings = new HashSet<>();
        review.setComments(reviewRatings);
        reviewRepository.save(review);

        mailService.sendEmail(
            System.getenv("MAIL_EMAIL"),
            "NEW REVIEW REQUESTED",
            user.get().getEmail() + " requested a review for " + request.getResource(),
            false,
            false
        );
    }

    public void notifyReviewers(Long id, String message) {
        Review review = reviewRepository.findById(id).get();
        List<String> emails = review
            .getComments()
            .stream()
            .map(reviewRating -> reviewRating.getUser().getEmail())
            .collect(Collectors.toList());
        emails.stream().forEach(email -> mailService.sendEmail(email, "Review Updated", message, false, false));
    }
}
